#!/bin/bash
unset OBF PASSWORD OUTPUTDIR

args=`getopt U:w:d:s: $*`
if test $? != 0
	then
	echo " "
	echo " Usage is identify_differently_sorted_buddies.sh [-U username] [-w password] [-d /path/to/output] [-s upgradeschema] "
	echo " Use database administrator account for user name "
	echo " "
	echo " Script to identify non-identically sorted buddy projections for a Vertica database"
	echo " Execute the DDL in /path/to/output/Out_*/deploy before upgrading to version 9.1"
	echo " "
	echo " Options: "
	echo " -d     path to generate output scripts  "
	echo " -s     schema name to store intermediate results "
	echo " "
	exit 1
fi

set -- $args

OUTPUTDIR="/tmp"
UPGRADESCHEMA="upgrade_vertica"

for i
do
	case "$i" in
		-U) shift;USER=$1;shift;;
		-w) shift;PASSWORD=$1;shift;;
		-d) shift;OUTPUTDIR=$1;shift;;
		-s) shift;UPGRADESCHEMA=$1;shift;;
	esac
done

# no USER environment varible set, and no username provided
# we should error out
if [ "X$USER" = "X" ]; then
    echo " "
    echo " Usage is identify_differently_sorted_buddies.sh [-U username] [-w password] [-d /path/to/output] [-s upgradeschema] "
    echo " Use database administrator account for user name "
    echo " "
    echo " Script to identify non-identically sorted buddy projections for a Vertica database"
    echo " Execute the DDL in /path/to/output/Out_*/deploy before upgrading to version 9.1"
    echo " "
    echo " Options: "
    echo " -d     path to generate output scripts  "
    echo " -s     schema name to store intermediate results "
    echo " "
    exit 1
fi

# if DATE is set in environment, use that otherwise default to timestamp
if [ "$DATE" = "" ] ; then
	DATE=`date +%Y%m%d%H%M%S`
fi

rm -rf $OUTPUTDIR/Out_*

OUT=$OUTPUTDIR/Out_$DATE

if [ -d $OUT ]; then    
#    echo " $OUT exists; removing all files contained"
	rm -rf $OUT/*
else    
#   echo "Creating $OUT ...."
	mkdir $OUT
	if [ $? -ne 0 ]; then
		echo "************************************************************************************************************"
		echo "Failed to create output scripts folder "
		echo "************************************************************************************************************"
		exit 0
	fi
	chmod 777 $OUT
fi

if [ "$USER" != "" ]; then
	us="-U $USER" 
	# user specified, need password
	# get password from user if not supplied 
	if [ "$PASSWORD" != "" ]; then
		pw="-w $PASSWORD"
	else
		if [ "X$VSQL_BINARY" = "X" ]; then
			read -p "Database Password:" -s PASSWORD
			echo " "
		fi
		if [ "$PASSWORD" != "" ]; then
			pw="-w $PASSWORD"
        fi
    fi
fi

# vsql to use: if VSQL env is set, use that otherwise use naked 'vsql'
if [ "X$VSQL_BINARY" = "X" ]; then 
    VSQL_BINARY="/opt/vertica/bin/vsql"
fi

myvsql="$VSQL_BINARY -X -At $us $pw -c "

# Check version, if earlier than 8.x, don't proceed
curVersion=`$myvsql "SELECT substr(split_part(version(), ' v', 2), 1, 1);"`
if [ -z "$curVersion" ]; then
    echo "************************************************************************************************************"
    echo "Failed to check Vertica version "
    echo "Database down or wrong username / password ?"
    echo "Read readme file \"README\""
    echo "************************************************************************************************************"
    exit 0
elif [ "$curVersion" -le 7 ]; then
    echo "************************************************************************************************************"
    echo "Your current Vertica version is older than v8.x "
    echo "Upgrade to version 8.x or 9.0 and re-run the script"
    echo "************************************************************************************************************"
    exit 0
fi

numNodes=`$myvsql "SELECT count(*) FROM nodes WHERE node_type='PERMANENT';"`
curK=`$myvsql "SELECT current_fault_tolerance FROM system;"`
scalingFactor=`$myvsql "select scaling_factor from vs_elastic_cluster;"`

echo -e "\nNumber of Nodes=$numNodes"
echo -e "Current system K=$curK"
echo -e "Current system scaling factor=$scalingFactor"

# Check if k-safety is 0, if so, no need to do anything
if [ "$curK" -eq 0 ]; then
    echo "************************************************************************************************************"
    echo "Your current system K-safety is 0. "
    echo "If this setting is expected, no need to run this script"
    echo "If this setting is unexpected, set the right k-safety in your database and rerun the script again"
    echo "************************************************************************************************************"
    exit 0
fi

# Check if working schema already exists
upgradeSchemaExists=`$myvsql "SELECT count(*) FROM vs_schemata WHERE name ilike '$UPGRADESCHEMA';"`

if [ "$upgradeSchemaExists" -eq 1 ]; then
    echo "************************************************************************************************************"
    echo "Schema $UPGRADESCHEMA already exists. This script requires a temporary working schema"
    echo "Option 1: re-run the script with \"-s\" option to specify a different working schema"
    echo "Option 2: drop the schema and re-run the script"
    echo "************************************************************************************************************"
    exit 1
fi

TOTAL=0

upgrade_projections="$UPGRADESCHEMA.projections"
upgrade_tables="$UPGRADESCHEMA.tables"
upgrade_segments="$UPGRADESCHEMA.segments"
upgrade_projection_columns="$UPGRADESCHEMA.projection_columns"
upgrade_projection_columns_hash="$UPGRADESCHEMA.projection_columns_hash"
upgrade_projections_internal="$UPGRADESCHEMA.projections_internal"

$myvsql "CREATE SCHEMA $UPGRADESCHEMA;"
$myvsql "CREATE TABLE $upgrade_projection_columns AS SELECT table_column_id,column_expression,column_position,sort_position,projection_id FROM v_catalog.projection_columns ORDER BY projection_id UNSEGMENTED ALL NODES;"
$myvsql "CREATE TABLE $upgrade_projection_columns_hash AS SELECT /*+direct*/ hash(sum(hash(table_column_id,column_expression,column_position,sort_position)::numeric)) as signature, projection_id FROM $upgrade_projection_columns group by projection_id order by projection_id UNSEGMENTED ALL NODES;"
$myvsql "CREATE TABLE $upgrade_projections AS SELECT p.projection_schema, p.projection_name, p.projection_id, p.anchor_table_id, p.has_expressions, p.is_segmented, p.is_up_to_date, isnull(p.segment_range,'') not ilike 'explicit range%' as is_not_seg_range, p.is_prejoin, p.is_aggregate_projection, seg.\"offset\" as off_set, seg.iselastic, hash(segment_expression) as seg, pcol_hash.signature FROM v_catalog.projections p left join v_internal.vs_segments seg on p.projection_id = seg.proj join $upgrade_projection_columns_hash pcol_hash on p.projection_id = pcol_hash.projection_id where projection_schema <> '$UPGRADESCHEMA' order by anchor_table_id,seg UNSEGMENTED ALL NODES;"
$myvsql "CREATE TABLE $upgrade_tables AS SELECT table_id, table_schema, table_name FROM v_catalog.tables where table_schema <> '$UPGRADESCHEMA' order by table_id UNSEGMENTED ALL NODES;"
$myvsql "CREATE TABLE $upgrade_segments AS SELECT \"offset\"%$numNodes as off_set, proj, allsites, iselastic from v_internal.vs_segments order by proj unsegmented all nodes;"
$myvsql "CREATE TABLE $upgrade_projections_internal AS SELECT oid, site, sitename FROM v_internal.vs_projections where schemaname <> '$UPGRADESCHEMA' and seginfo = 0 order by oid UNSEGMENTED ALL NODES;"

# Check system tables
stashed_table_count=`$myvsql "SELECT count(*) FROM v_catalog.tables where table_schema = '$UPGRADESCHEMA'"`

if [ "$stashed_table_count" -ne "6" ]; then
    echo "************************************************************************************************************"
        echo "Schema $UPGRADESCHEMA is not populated successfully, some tables are missing"
        echo "Possible reason: if you see 'ERROR 4305: Out of system WOS memory during catalog SELECT ... '"
    echo "Solution: increase sysdata max memory by setting 'ALTER RESOURCE POOL sysdata maxmemorysize '<mem size>';"
        echo "************************************************************************************************************"
    exit 0
fi

# Check if any projections are not fully rebalanced
has_proj_unbalanced=`$myvsql "SELECT count(*) FROM $upgrade_segments where iselastic and allsites != $scalingFactor*$numNodes;"`

if [ "$has_proj_unbalanced" -ne "0" ]; then
    echo "************************************************************************************************************"
        echo "$has_proj_unbalanced projections are not fully segmented but not on all permanent nodes"
        echo "Possible reason: some nodes are newly added but not ran rebalance_cluster yet"
    echo "Solution 1: Run rebalance_cluster and rerun this script"
    echo "Solution 2: Mark newly added node as ephemeral and rerun this script"
        echo "************************************************************************************************************"
    $myvsql "drop schema $UPGRADESCHEMA cascade"
    exit 0
fi

# Analyze segmented projections (not expression projection)
echo -e "Analyzing segmented projection buddy groups"

unset projections_processed
declare -A projections_processed

rebuild_projection_names=()

for projs in `$myvsql "SELECT projection_schema,projection_name,projection_id,anchor_table_id,has_expressions,iselastic,off_set,seg,signature FROM 
    $upgrade_projections 
    WHERE is_segmented 
    AND is_up_to_date 
    AND is_not_seg_range 
    AND not is_prejoin 
    AND not is_aggregate_projection 
    ORDER BY anchor_table_id, projection_id;"`
do

#STARTTIME=$(date +%s%N)

    sch=`echo $projs| awk 'BEGIN {FS = "|"} {print $1}'`
    proj=`echo $projs| awk 'BEGIN {FS = "|"} {print $2}'`
    proj_id=`echo $projs| awk 'BEGIN {FS = "|"} {print $3}'`
    tbl_id=`echo $projs| awk 'BEGIN {FS = "|"} {print $4}'`
    expr_proj=`echo $projs| awk 'BEGIN {FS = "|"} {print $5}'`
    iselastic=`echo $projs| awk 'BEGIN {FS = "|"} {print $6}'`
    offset=`echo $projs| awk 'BEGIN {FS = "|"} {print $7}'`
    seg=`echo $projs| awk 'BEGIN {FS = "|"} {print $8}'`
    signature=`echo $projs| awk 'BEGIN {FS = "|"} {print $9}'`

    proj_buds=($proj_id)
    echo -n "Projection $sch.$proj ... "
        
    if [[ ${projections_processed[$proj_id]} ]]; then
    echo -e "already processed"
    else
    ################################################
    # [Run query in Vertica] Get identical buddies #
    ################################################
    get_buddies_qry="select projection_id,projection_name from $upgrade_projections where (anchor_table_id = $tbl_id) and is_up_to_date and (seg = '$seg') and (off_set <> $offset) and (signature = $signature) and (has_expressions = '$expr_proj');"
    allbuddies=`$myvsql "$get_buddies_qry"`

    skip_buds=()
    for buddy in $allbuddies
    do
        buddy_id=`echo $buddy| awk 'BEGIN {FS = "|"} {print $1}'`
        buddy_name=`echo $buddy| awk 'BEGIN {FS = "|"} {print $2}'`
        proj_buds+=($buddy_id)
        skip_buds+=("$tbl_id=$sch.$buddy_name")
        projections_processed[$buddy_id]=1
    done

    ###############################################################
    # [Run query in Vertica] Get distinct offsets of real buddies #
    ###############################################################
    in_list=$(IFS=,; echo "(${proj_buds[*]})")
    offset_result=`$myvsql "select distinct off_set from $upgrade_segments where proj in $in_list order by 1;"`

    offsets=()
    for offs in $offset_result
    do
        offsets+=($offs)
    done
    buddy_group_size=${#offsets[@]}

    if [ "$curK" -ge "$buddy_group_size" ]; then
        offset_str=$(IFS=" "; echo "${offsets[*]}")
        echo -e "unsafe"
        if [ "$expr_proj" = "f" ] && [ "$iselastic" = "t" ]; then
            echo -e "$sch.$proj.offsets=$offset_str" >> $OUT/segmented_projections
        else
            echo -e "$sch.$proj.offsets=$offset_str" >> $OUT/rebuild_projection
                rebuild_projection_names+=("$tbl_id=$sch.$proj")
            for i in ${skip_buds[@]}; do rebuild_projection_names+=($i); done
        fi
    else
        echo -e "ok"
    fi
    fi

#ENDTIME=$(date +%s%N)
#ELAPSED=$((($ENDTIME - $STARTTIME)/1000000))
#TOTAL=$(($TOTAL + $ELAPSED))
#echo "[Process segment: Elapsed = $ELAPSED; TOTAL = $TOTAL]"
done

# Analyze unsegmented projections
echo -e "\n\nAnalyzing unsegmented projection buddy groups"

unset projections_processed
declare -A projections_processed

nodes=()
for n in `$myvsql "select node_name from nodes where node_type='PERMANENT' order by node_name;"`
do
    nodes+=($n)
done

for projs in `$myvsql "select projection_schema,projection_name,projection_id,anchor_table_id,has_expressions,signature from 
    $upgrade_projections 
    where not is_segmented 
    and is_up_to_date and is_not_seg_range
    and not is_prejoin 
    and not is_aggregate_projection 
    order by anchor_table_id, projection_id;"`
do
    sch=`echo $projs| awk 'BEGIN {FS = "|"} {print $1}'`
    proj=`echo $projs| awk 'BEGIN {FS = "|"} {print $2}'`
    proj_id=`echo $projs| awk 'BEGIN {FS = "|"} {print $3}'`
    tbl_id=`echo $projs| awk 'BEGIN {FS = "|"} {print $4}'`
    expr_proj=`echo $projs| awk 'BEGIN {FS = "|"} {print $5}'`
    signature=`echo $projs| awk 'BEGIN {FS = "|"} {print $6}'`
    proj_buds=($proj_id)
    echo -n "Projection $sch.$proj ... "

    if [[ ${projections_processed[$proj_id]} ]]; then
    echo -e "already processed"
    else
    is_broadcast_proj=`$myvsql "select (select (select count (distinct site) from $upgrade_projections_internal where oid=$proj_id) = $numNodes);"`
    if [ "$is_broadcast_proj" = "t" ]; then
        echo -e "ok"
        projections_processed[$proj_id]=1
    else
        ################################################
        # [Run query in Vertica] Get identical buddies #
        ################################################
        proj_buds=($proj_id)
        get_buddies_qry="select projection_id,projection_name from $upgrade_projections 
        where (projection_id <> $proj_id) and (signature = $signature)
        and not is_segmented and is_not_seg_range
        and not is_prejoin and is_up_to_date and (anchor_table_id = $tbl_id) and (has_expressions = '$expr_proj');"

        skip_buds=()

        for buddy in `$myvsql "$get_buddies_qry"`
        do
        buddy_id=`echo $buddy| awk 'BEGIN {FS = "|"} {print $1}'`
        buddy_name=`echo $buddy| awk 'BEGIN {FS = "|"} {print $2}'`
        proj_buds+=($buddy_id)
        skip_buds+=("$tbl_id=$sch.$buddy_name")
        projections_processed[$buddy_id]=1
        done
        
        #################################################################
        # [Run query in Vertica] Get distinct node sets of real buddies #
        #################################################################
        in_list=$(IFS=,; echo "(${proj_buds[*]})")
        satisfied_nodes=()
        for n in `$myvsql "select distinct sitename from $upgrade_projections_internal where oid in $in_list order by sitename;"`
        do
        satisfied_nodes+=($n)
        done
        satisfied_nodes_size=${#satisfied_nodes[@]}
    
        if [ "$satisfied_nodes_size" -lt "$numNodes" ]; then
        satisfied_nodes_str=$(IFS=" "; echo "${satisfied_nodes[*]}")
        echo -e "unsafe"
            if [ "$expr_proj" = "f" ]; then
            echo -e "$sch.$proj.satisfied_nodes=$satisfied_nodes_str" >> $OUT/unsegmented_projections
        else
            echo -e "$sch.$proj.satisfied_nodes=$satisfied_nodes_str" >> $OUT/unsegmented_rebuild_projection
            rebuild_projection_names+=("$tbl_id=$sch.$proj")
                    for i in ${skip_buds[@]}; do rebuild_projection_names+=($i); done
        fi
        else
        echo -e "ok"
        fi
    fi
    fi
done

#echo "Finished exporting 

unusual_projs=`$myvsql "select projection_schema||'.'||projection_name from $upgrade_projections where is_prejoin or (is_segmented and not is_not_seg_range);"`

if [ -n "$unusual_projs" ]; then
    echo -e "\n\nThe database contains prejoin projections and/or range segmented projections which will be deprecated in the next release"
    echo -e "Manually validate whether the sort orders of the following buddy groups are identical prior to upgrade"

    for i in "${unusual_projs[@]}"
    do
    echo -e "$i"
    done
fi

nonuptodate_projs=`$myvsql "select distinct projection_schema||'.'||projection_name,anchor_table_id from $upgrade_projections where not is_up_to_date order by anchor_table_id;"`

if [ -n "$nonuptodate_projs" ]; then
    echo -e "\n\nThe database contains non-up-to-date projections that need to be dropped prior to upgrade"
    echo -e "Execute the DROP PROJECTION statements in the deployment script"
    echo -e "The following projections will be dropped"

    previous_tbl_id=0
    drop_stmt=''
    readarray -t rows <<<"$nonuptodate_projs"

    for row in "${rows[@]}"
    do
    proj=`echo $row| awk 'BEGIN {FS = "|"} {print $1}'`
    tbl_id=`echo $row| awk 'BEGIN {FS = "|"} {print $2}'`

    echo -e "$proj"
    if [[ "$tbl_id" == "$previous_tbl_id" ]]; then
        drop_stmt+=", $proj"
    else
        if [ "${#drop_stmt}" -ne 0 ]; then
        drop_stmt+=" CASCADE;"
        echo -e "$drop_stmt" >> $OUT/deploy
        fi
        drop_stmt="DROP PROJECTION $proj"
        previous_tbl_id=$tbl_id
    fi
    done
    if [ "${#drop_stmt}" -ne 0 ]; then
    drop_stmt+=" CASCADE;"
    echo -e "$drop_stmt" >> $OUT/deploy
    fi
fi

# no adjustment required
if [ ! -e $OUT/segmented_projections ] && [ ! -e $OUT/unsegmented_projections ] && [ ! -e $OUT/rebuild_projection ] && [ ! -e $OUT/unsegmented_rebuild_projection ] && [ ! -n "$nonuptodate_projs" ]; then
    $myvsql "drop schema $UPGRADESCHEMA cascade;"

    echo "************************************************************************************************************"
    echo "Congratulations! No unsafe projections detected. Upgrade to 9.1 should succeed"
    echo "************************************************************************************************************"

    exit 0
fi

declare -A used_names

unique_name() {
    sch=$1
    proj=$2
    un="$3"
    count=1

    while [ true ] ; do
    used=true
    while [ $used = true ] ; do
        new_proj_name=$proj"_v"$count

        if [[ ${used_names[$new_proj_name]} ]]; then
        let "count+=1"
        else
        used=false
        fi
    done

    duplicate=`$myvsql "select count(*) from $upgrade_projections where projection_schema='$sch' and projection_name='$new_proj_name';"`
    if [ $duplicate -eq 0 ]; then
        used_names[$new_proj_name]=1
        return
    else
        let "count+=1"
    fi
    done
}

generate_projection_ddl() {
    sch=$1
    proj=$2
    offset_or_node=$3
    orig_ddl=$4
    segmented=$5
    is_rebuild=$6

    unique_name $sch $proj
    output_proj_ddl=`sed "s/$sch.$proj/$sch.$new_proj_name/" <<< $orig_ddl`
    if [ "$segmented" = true ]; then
    if [ "$is_rebuild" = true ]; then
        output_proj_ddl=`sed "s/OFFSET [0-9]\+//" <<< $output_proj_ddl`
        output_proj_ddl=`sed "s/NODES [^;]\+/ALL NODES /" <<< $output_proj_ddl`
        output_proj_ddl=`sed "s/;/KSAFE;/" <<< $output_proj_ddl`
    else
        output_proj_ddl=`sed "s/OFFSET [0-9]\+//" <<< $output_proj_ddl`
        output_proj_ddl=`sed "s/;/OFFSET $offset_or_node;/" <<< $output_proj_ddl`
    fi
    else
    if [ "$is_rebuild" = true ]; then
        output_proj_ddl=`sed "s/UNSEGMENTED NODE .*;/UNSEGMENTED ALL NODES;/" <<< $output_proj_ddl`
    else
        output_proj_ddl=`sed "s/UNSEGMENTED NODE .*;/UNSEGMENTED NODE $offset_or_node;/" <<< $output_proj_ddl`
    fi
    fi
    echo -e "$output_proj_ddl" >> $OUT/deploy
}

# Generate DDLs for segmented projections 
if [ -e $OUT/segmented_projections ]; then

    echo -e "\n\nProcessing deployment of segmented projections"

    while IFS='.' read -r sch proj offsets
    do
    proj_ddl=`$myvsql "select export_objects('','$sch.$proj');"`
    proj_ddl=`sed 's/SELECT MARK_DESIGN_KSAFE([0-9]\+);//' <<< $proj_ddl`
    offsets=${offsets:8}
    read -a offset_arr <<<$offsets

    associative_offset_array=()
    for i in "${offset_arr[@]}"
    do
        associative_offset_array[$i % $numNodes]=1
    done

    for (( i=0; i<=$curK; i++ ))
    do
        if [[ ! ${associative_offset_array[$i]} ]]; then
        generate_projection_ddl $sch $proj $i "$proj_ddl" true false
        fi
    done
    done <$OUT/segmented_projections
fi

# Generate DDLs for unsegmented projections 
if [ -e $OUT/unsegmented_projections ]; then

    echo -e "\n\nProcessing deployment of unsegmented projections"

    while IFS='.' read -r sch proj sat_nodes
    do
    proj_ddl=`$myvsql "select export_objects('','$sch.$proj');"`
    proj_ddl=`sed 's/SELECT MARK_DESIGN_KSAFE([0-9]\+);//' <<< $proj_ddl`
    sat_nodes=${sat_nodes:16}
    satisfied_nodes=()
    read -a satisfied_nodes <<<$sat_nodes

    remaining_nodes=(`echo ${nodes[@]} ${satisfied_nodes[@]} | tr ' ' '\n' | sort | uniq -u `) 
    
    for i in "${remaining_nodes[@]}"
    do
        generate_projection_ddl $sch $proj $i "$proj_ddl" false false
    done
    done <$OUT/unsegmented_projections
fi

# Generate DDLs for expression / in-elastic projections 
if [ -e $OUT/rebuild_projection ]; then

    echo -e "\n\nProcessing deployment of expression / inelastic projections"

    while IFS='.' read -r sch proj offsets
    do
    proj_ddl=`$myvsql "select export_objects('','$sch.$proj');"`
    proj_ddl=`sed 's/SELECT MARK_DESIGN_KSAFE([0-9]\+);//' <<< $proj_ddl`

    generate_projection_ddl $sch $proj $i "$proj_ddl" true true
    done <$OUT/rebuild_projection
fi

# Generate DDLs for unsegmented expression projections 
if [ -e $OUT/unsegmented_rebuild_projection ]; then

    echo -e "\n\nProcessing deployment of unsegmented expression / inelastic projections"

    while IFS='.' read -r sch proj offsets
    do
    proj_ddl=`$myvsql "select export_objects('','$sch.$proj');"`
    proj_ddl=`sed 's/SELECT MARK_DESIGN_KSAFE([0-9]\+);//' <<< $proj_ddl`

    generate_projection_ddl $sch $proj $i "$proj_ddl" false true
    done <$OUT/unsegmented_rebuild_projection
fi

projection_oids=()

if [ -e $OUT/segmented_projections ]; then

    while IFS='.' read -r sch proj offsets
    do
    projection_oids+=(`$myvsql "select projection_id from $upgrade_projections where projection_schema='$sch' and projection_name='$proj';"`)
    done <$OUT/segmented_projections
fi

if [ -e $OUT/unsegmented_projections ]; then

    while IFS='.' read -r sch proj sat_nodes
    do
    projection_oids+=(`$myvsql "select projection_id from $upgrade_projections where projection_schema='$sch' and projection_name='$proj';"`)
    done <$OUT/unsegmented_projections
fi

if [ -e $OUT/rebuild_projection ]; then

    while IFS='.' read -r sch proj offsets
    do
    projection_oids+=(`$myvsql "select projection_id from $upgrade_projections where projection_schema='$sch' and projection_name='$proj';"`)
    done <$OUT/rebuild_projection
fi

if [ -e $OUT/unsegmented_rebuild_projection ]; then

    while IFS='.' read -r sch proj sat_nodes
    do
    projection_oids+=(`$myvsql "select projection_id from $upgrade_projections where projection_schema='$sch' and projection_name='$proj';"`)
    done <$OUT/unsegmented_rebuild_projection
fi

projection_oid_list=$(IFS=,; echo "(${projection_oids[*]})")

if [ -n "$projection_oids" ]; then

    table_names=()

    for row in `$myvsql "select distinct table_schema, table_name from $upgrade_projections, $upgrade_tables where anchor_table_id=table_id and projection_id in $projection_oid_list order by 1,2;"`
    do
        sch=`echo $row| awk 'BEGIN {FS = "|"} {print $1}'`
        tbl=`echo $row| awk 'BEGIN {FS = "|"} {print $2}'`
        table_names+=($sch.$tbl)
    done

    table_name_list=$(IFS=,; echo "('${table_names[*]}')")
    echo -e "SELECT REFRESH $table_name_list;" >> $OUT/deploy
fi

# For expression projection, drop existing ones because we already create new ksafe groups
if [[ ${rebuild_projection_names[@]} ]]; then

    echo -e "\n\nThe database contains existing unsafe expression / inelastic projections that need to be dropped prior to upgrade"
    echo -e "Execute the DROP PROJECTION statements in the deployment script"
    echo -e "The following projections will be dropped"

    last_tbl_oid=0
    proj_list=()

    # print out make_ahm_now because drop projection may fail
    echo -e "SELECT make_ahm_now();" >> $OUT/deploy

    for i in ${rebuild_projection_names[@]}
    do
        tbl_oid=`echo $i| awk 'BEGIN {FS = "="} {print $1}'`
        proj_name=`echo $i| awk 'BEGIN {FS = "="} {print $2}'`
        echo -e "$proj_name"
        # new table, flush existing items if there is any
        if [ "$tbl_oid" -ne "$last_tbl_oid" ]; then
            if [ -n "$proj_list" ]; then
                rebuild_projection_name_list=$(IFS=,; echo "${proj_list[*]}")
                echo -e "DROP PROJECTION $rebuild_projection_name_list;" >> $OUT/deploy
            fi
            last_tbl_oid=$tbl_oid
            unset proj_list
            proj_list=()
        fi
        proj_list+=("$proj_name")
    done

    # final flush if there is any
    if [ -n "$proj_list" ]; then
        rebuild_projection_name_list=$(IFS=,; echo "${proj_list[*]}")
        echo -e "DROP PROJECTION $rebuild_projection_name_list;" >> $OUT/deploy
    fi
fi

$myvsql "drop schema $UPGRADESCHEMA cascade"

# show to-do's

echo "************************************************************************************************************"
echo "Caution! Detected unsafe projections. Upgrade to 9.1 will fail. Please follow the following instructions"
echo "1) Read readme file \"README\""
echo "2) Review and run generated fixing script \"$OUT/deploy\". This may take a while"
echo "3) Rerun this script again to make sure no unsafe projections are detected"
echo "4) If unsafe projection cannot be eliminated completely, please contact Vertica"
echo "************************************************************************************************************"

exit 0
